GREEN = "\033[92m"
ORANGE = "\033[93m"
RED = "\033[91m"
GRAY = "\033[90m"
BOLD = '\033[1m'
END = "\033[0m"


def _format(style, text):
    return f"{style}{text}{END}"


def green(text):
    return _format(GREEN, text)


def orange(text):
    return _format(ORANGE, text)


def red(text):
    return _format(RED, text)


def gray(text):
    return _format(GRAY, text)


def bold(text):
    return _format(BOLD, text)



import os
import sys
from abc import ABC, abstractmethod
from collections import defaultdict

from linter.helpers.exceptions import LintingError, exception_handler
from linter.helpers.format import gray, bold, green


class LintController:
    errors = defaultdict(list)
    warnings = defaultdict(list)

    def output(self):
        sys.excepthook = exception_handler

        if self.errors:
            raise LintingError(self.errors)
        else:
            print(bold(green("\n🤖 HTML looks good to me!\n")))

    def __init__(self, linters):
        base_directory = os.getcwd()
        templates_directory = f"{base_directory}/templates"

        print('\n')
        print(templates_directory)
        print('\n')

        # for root, _, files in os.walk(templates_directory):
        #     for file_name in files:
        #         if file_name.endswith(".html"):
        #             with open(f"{root}/{file_name}") as f:
        #                 lines = f.readlines()
        #                 text = '\n'.join(lines)
        #
        #                 folders_we_care_about = str(root)[len(templates_directory) + 1:] + '/'
        #                 file_name = file_name
        #                 key = folders_we_care_about + file_name
        #
        #                 for linter in linters:
        #                     linter = linter(key, base_directory, lines, text)
        #                     linter.lint()
        #
        #                     for key, value in linter.errors.items():
        #                         if isinstance(value, str):
        #                             value = [value]
        #
        #                         self.errors[key] = self.errors[key] + value
        #
        #                     for key, value in linter.warnings.items():
        #                         if isinstance(value, str):
        #                             value = [value]
        #
        #                         self.warnings[key] = self.warnings[key] + value


class Linter(ABC):
    def __init__(self, file, base_directory, lines, text):
        self.file = file
        self.base_directory = base_directory
        self.lines = lines
        self.text = text
        self.errors = defaultdict(list)
        self.warnings = defaultdict(list)

    @abstractmethod
    def lint(self):
        return self.errors, self.warnings

    def append_error(self, error, line_number: int = -1, more_info=""):
        if line_number > 0:
            line_number = gray(" (line " + str(line_number) + ")")
        else:
            line_number = ""

        if more_info:
            more_info = "\n    " + gray(more_info)

        self.errors[self.file].append(error + line_number + more_info)

    def append_warning(self, warning, _, __):
        self.warnings[self.file].append(warning)



import os
import re

from bs4 import BeautifulSoup

from linter.helpers.list_helpers import index_of_string_in_list
from linter.linters.base import Linter


class CheckHeadings(Linter):
    """
    All top level headings should be of type H1, not type H2
    """

    def lint(self):
        for i in range(len(self.lines)):
            if '<h2' in self.lines[i].strip() and 'class="govuk-heading-l"' in self.lines[i].strip():
                self.append_error("The page heading should be H1", i + 1, self.lines[i].strip())


class FindHardcodedStrings(Linter):
    """
    Finds hardcoded strings in HTML
    """

    def lint(self):
        soup = BeautifulSoup(self.text, "html.parser")
        result = soup.stripped_strings

        for i in result:
            if not i.startswith("{{") and not i.startswith("{%"):
                self.append_error("Use LITE Content Library instead of hardcoding strings",
                                  index_of_string_in_list(i, self.lines), i)


class TagsSpacing(Linter):
    """
    Tags should have a space after them eg {{ hello }} instead of {{hello}}
    or {% hello %} instead of {%hello%}
    """

    def _lint(self, char, char2):
        tags = re.findall("{" + char + "(\S*?)" + char2 + "}", self.text)
        for item in tags:
            self.append_error("Tag contents should have a surrounding space eg "
                              "{" + char + " " + item + " " + char2 + "} instead of {" + char + item + char2 + "}",
                              index_of_string_in_list("{" + char + item + char2 + "}", self.lines))

    def lint(self):
        self._lint("{", "}")
        self._lint("%", "%")


class CheckFilenames(Linter):
    """
    HTML files and folders should use dashes instead of underscores in its filename
    """

    def lint(self):
        if "_" in self.file:
            self.append_error("Use dashes instead of underscores in filenames and folders, like so: " +
                              self.file.replace("_", "-"))


class EnforceBlockTitle(Linter):
    """
    All pages should have {% block title %}
    """

    def lint(self):
        if "{% block title %}" not in self.text and "includes/" not in self.file and "tabs/" not in self.file:
            self.append_error("Missing page title - use {% block title %}Title goes here{% endblock %} to set it")


class UnusedFiles(Linter):
    """
    Find unused HTML files
    """

    def _find_file(self, look_for_file_name):
        for root, _, files in os.walk(self.base_directory):
            for file_name in files:
                if file_name.endswith(".html") or file_name.endswith(".py"):
                    with open(f"{root}/{file_name}") as f:
                        if look_for_file_name in f.read():
                            return True

        return False

    def lint(self):
        if self.file.endswith(".html") and not self.file.endswith("404.html") and not self.file.endswith("500.html"):
            if not self._find_file(self.file):
                self.append_error("⚠️  This file isn't being used anywhere")


class PreventInlineStyles(Linter):
    """
    Prevent inline styles
    """

    def lint(self):
        for i in range(len(self.lines)):
            if 'style=' in self.lines[i].strip() and "display:none;visibility:hidden" not in self.lines[i].strip():
                self.append_error("Don't use inline styles", i + 1, self.lines[i].strip())


class OneLoadBlock(Linter):
    """
    Each HTML file should have at most one {% load %} block
    """

    def lint(self):
        load_blocks = [x for x in self.lines if "{% load" in x]
        if len(load_blocks) > 1:
            load_blocks.pop(0)
            initial_index = 0
            for block in load_blocks:
                index = self.lines.index(block, initial_index) + 1
                self.append_error("More than one {% load %} block", index, block)
                initial_index = index


class AllTagsUsed(Linter):
    """
    Ensure each template tag imported in {% load %} is actually used
    """

    def lint(self):
        load_blocks = [x for x in self.lines if "{% load" in x]
        if load_blocks:
            load_block = load_blocks.pop(0)
            template_tags = load_block[7:-3].strip().split(" ")

            for template_tag in template_tags:
                if template_tag == "sass_tags":
                    continue
                text = self.text.replace(load_block, "")
                if template_tag not in text:
                    self.append_error(f"Template tag '{template_tag}' not used in file",
                                      index_of_string_in_list(load_block, self.lines),
                                      load_block)


class EnsureAllIDsAreUnique(Linter):
    """
    Ensure all IDs used in a file are unique to that file
    """

    def lint(self):
        reg = "id=\"(.*?)\""
        ids = re.findall(reg, self.text)
        ids_completed = []
        for id in ids:
            ids_completed.append(id)
            if ids_completed.count(id) == 2:
                self.append_error("IDs should be unique",
                                  0,
                                  "id=\"" + id + "\"")


class BlockOrdering(Linter):
    """
    {% load %} block should be after {% extends %}, but before first {% block %}
    """

    def lint(self):
        load_blocks = [x for x in self.lines if "{% load" in x]
        try:
            extend_block = self.text.index("{% extends")
            load_block = self.text.index("{% load")
            block_block = self.text.index("{% block")

            if load_block < extend_block or load_block > block_block:
                self.append_error("{% load %} block should be after {% extends %}, but before first {% block %}",
                                  self.lines.index(load_blocks[0]) + 1)
        except ValueError:
            pass


class PreventTwoPaneOverwite(Linter):
    """
    Ensure that body isn't overwritten in two-pane pages
    """

    def lint(self):
        if "two-pane.html" in self.text and "{% block body %}" in self.text:
            self.append_error("Extend base.html instead of two-pane.html if you're just going to override body")


class EnforceSingleQuotesInBlocks(Linter):
    """
    {%%} blocks should have single quotes, not double
    """

    def lint(self):
        blocks = re.findall("{% (.*) %}", self.text)
        for item in blocks:
            if "\"" in item:
                self.append_error("{%%} Blocks should use single quotes not double",
                                  index_of_string_in_list(item, self.lines))


class BlocksSpacing(Linter):
    """
    Ensure blocks are spaced correctly
    """

    def lint(self):
        # Blocks should NOT have an empty line below them
        # eg
        # {% block javascript %}
        # 	DO STUFF HERE!
        # {% endblock %}
        for i in range(len(self.lines)):
            if self.lines[i].startswith("{% block") and not self.lines[i].rstrip().endswith("endblock %}"):
                try:
                    if self.lines[i + 1].strip() == "":
                        self.append_error("Blocks should not have an empty line after them",
                                          i + 1,
                                          self.lines[i].strip() + '\n    ' + self.lines[i + 1].strip())
                except IndexError:
                    pass

        # End Blocks should NOT be proceeded with an empty line
        # eg
        # {% block javascript %}
        # 	DO STUFF HERE!
        # {% endblock %}
        for i in range(len(self.lines)):
            if self.lines[i].startswith("{% endblock %}"):
                try:
                    if self.lines[i - 1].strip() == "":
                        self.append_error("End blocks should not have an empty line before them",
                                          i - 1,
                                          self.lines[i].strip() + '\n    ' + self.lines[i - 1].strip())
                except IndexError:
                    pass

        # End Blocks should be succeeded by an empty line
        # eg
        #   ...
        # {% endblock %}
        #
        # {% block javascript %}
        # 	...
        # {% endblock %}
        for i in range(len(self.lines)):
            if self.lines[i].rstrip() == "{% endblock %}":
                try:
                    if self.lines[i + 1].strip() != "":
                        self.append_error("End blocks should be succeeded by an empty line",
                                          i + 1,
                                          self.lines[i].strip() + '\n    ' + self.lines[i + 1].strip())
                except IndexError:
                    pass


class TabsOverSpaces(Linter):
    def lint(self):
        for line in self.lines:
            indent = abs(len(line.lstrip()) - len(line))
            substring = line[:indent]
            if " " in substring:
                self.append_error("We use tabs over spaces", index_of_string_in_list(line, self.lines))



import os

import setuptools

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

_INSTALL_REQUIRES = [
    'beautifulsoup4>=4.8.2',
]

setuptools.setup(
    name='lite-linter',
    version='0.0.8',
    author="Jan Faracik",
    author_email="",
    description="Linting for LITE",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/uktrade/lite-linter",
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'lint = linter.lint:main',
        ],
    },
    install_requires=_INSTALL_REQUIRES,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from linter.helpers.format import bold, red


def exception_handler(_, exception, __):
    """
    Override the default exception handler to hide pointless stacktrace
    and hide exception name for linting errors (error text is satisfactory)
    """
    print(exception)


class LintingError(Exception):
    def __init__(self, errors2: dict):
        count = 0

        for x in errors2.values():
            for _ in x:
                count = count + 1

        if count == 1:
            error_title = "There is 1 linting error"
        else:
            error_title = f"🔥 There are {count} linting errors"

        formatted_errors = '\n\n'.join([(x + '\n' +
                                         '\n'.join([('- ' + i) for i in y])) for x, y in
                                        errors2.items()]) + '\n'

        super().__init__(formatted_errors + '\n' + bold(red(error_title)) + '\n')



import sys

from linter.linters import headings
from linter.linters.base import LintController


def main():
    LintController(
        linters=[
            headings.CheckHeadings,
            headings.TagsSpacing,
            headings.UnusedFiles,
            headings.PreventInlineStyles,
            headings.OneLoadBlock,
            headings.AllTagsUsed,
            headings.EnsureAllIDsAreUnique,
            headings.BlockOrdering,
            headings.BlocksSpacing,
            headings.PreventTwoPaneOverwite,
            headings.CheckFilenames,
            headings.EnforceBlockTitle,
            headings.FindHardcodedStrings,
            headings.EnforceSingleQuotesInBlocks,
            headings.TabsOverSpaces,
        ]
    ).output()


if __name__ == "__main__":
    sys.exit(main())
